@@ -1,10 +1,12 @@
 use crate::checkout::Checkout;
 use crate::context::Context;
 use crate::error::new_application_error;
+use crate::error::new_invalid_state_error;
 use crate::error::new_not_found_error;
 use crate::error::Error;
 use crate::internal_context::InternalContext;
 use crate::payment::cancel as cancel_payment;
+use crate::payment::handlers::step_handler;
 use crate::payment::Initiator;
 use crate::payment::Payment;
 use crate::payment::PaymentProcessor;
@@ -35,6 +37,38 @@ pub async fn get_handler<C: Context + Send>(
     Ok(od)
 }
 
+pub async fn step_payment_handler<C: Context + Send>(
+    internal_ctx: &InternalContext,
+    ctx: &mut C,
+    order_id: &str,
+    args: <C as PaymentProcessor>::ProcessArgs,
+) -> Result<
+    Order<Payment<<C as PaymentProcessor>::Data, <C as PaymentProcessor>::ProcessArgs>>,
+    Error,
+> {
+    let maybe_od = ctx.get_order_for_update(order_id).await?;
+    if maybe_od.is_none() {
+        return Err(new_not_found_error(
+            "the requested order could not be found",
+        ));
+    }
+
+    let mut od = maybe_od.unwrap();
+
+    match od.state {
+        OrderState::PaymentInProgress => {
+            let payment_id = od.payment_id.as_ref().unwrap();
+            step_handler(internal_ctx, ctx, payment_id, args).await?;
+
+            od.populate_associations(ctx).await?;
+            Ok(od)
+        }
+        _ => Err(new_invalid_state_error(
+            "order must be in the PaymentInProgress state",
+        )),
+    }
+}
+
 pub async fn confirm_handler<C: Context + Send>(
     internal_ctx: &InternalContext,
     ctx: &mut C,
@@ -77,21 +111,24 @@ pub async fn cancel_handler<C: Context + Send>(
 
     let mut od = maybe_od.unwrap();
 
-    if let OrderState::PaymentInProgress = od.state {
-        if maybe_payment.is_none() {
-            return Err(new_not_found_error(
-                "the requested payment could not be found",
-            ));
-        }
+    match od.state {
+        OrderState::Cancelled => {}
+        OrderState::PaymentInProgress => {
+            let mut payment = maybe_payment.unwrap();
 
-        let mut payment = maybe_payment.unwrap();
-        cancel_payment(internal_ctx, ctx, &mut payment, Initiator::Order(&mut od)).await?;
-    } else {
-        od.cancel(internal_ctx, ctx).await?;
+            // check that the order payment hasn't changed
+            if &payment.id != od.payment_id.as_ref().unwrap() {
+                return Err(new_application_error("STALE_PAYMENT", "the associated payment was changed by another process while this transaction was in progress. please try again."));
+            }
+
+            cancel_payment(internal_ctx, ctx, &mut payment, Initiator::Order(&mut od)).await?;
+        }
+        _ => {
+            od.cancel(internal_ctx, ctx).await?;
+        }
     }
 
     od.populate_associations(ctx).await?;
-
     Ok(od)
 }
 
@@ -152,12 +189,7 @@ async fn get_order_payment_for_update<C: Context + Send>(
     Error,
 > {
     let maybe_od: Option<
-        Order<
-            Payment<
-                <C as PaymentProcessor>::Data,
-                <C as PaymentProcessor>::ProcessArgs,
-            >,
-        >,
+        Order<Payment<<C as PaymentProcessor>::Data, <C as PaymentProcessor>::ProcessArgs>>,
     > = ctx.get_order(order_id).await?;
     if maybe_od.is_none() {
         return Err(new_not_found_error(
